#
# Commands for the build environment
#
MIG =  $(NEXT_ROOT)/usr/bin/mig

MD=     /usr/bin/md

RM = /bin/rm -f
CP = /bin/cp
LN = /bin/ln -s
CAT = /bin/cat
MKDIR = /bin/mkdir -p

TAR = /usr/bin/gnutar
STRIP = /usr/bin/strip
LIPO = /usr/bin/lipo

BASENAME = /usr/bin/basename
RELPATH = $(NEXT_ROOT)/usr/local/bin/relpath
SEG_HACK = $(NEXT_ROOT)/usr/local/bin/seg_hack

UNIFDEF   = /usr/bin/unifdef
DECOMMENT = /usr/local/bin/decomment

